@@ -4,11 +4,17 @@ Low level socket utilities.
 Pyro - Python Remote Objects.  Copyright by Irmen de Jong (irmen@razorvine.net).
 """
 
+import os
 import socket
 import errno
 import time
 import sys
 import select
+import weakref
+try:
+    import ssl
+except ImportError:
+    ssl = None
 from Pyro4.configuration import config
 from Pyro4.errors import CommunicationError, TimeoutError, ConnectionClosedError
 
@@ -132,13 +138,14 @@ def receiveData(sock, size):
         retrydelay = 0.0
         msglen = 0
         chunks = []
-        if config.USE_MSG_WAITALL:
+        if config.USE_MSG_WAITALL and not hasattr(sock, "getpeercert"):
             # waitall is very convenient and if a socket error occurs,
             # we can assume the receive has failed. No need for a loop,
             # unless it is a retryable error.
             # Some systems have an erratic MSG_WAITALL and sometimes still return
             # less bytes than asked. In that case, we drop down into the normal
             # receive loop to finish the task.
+            # Also note that on SSL sockets, you cannot use MSG_WAITALL (or any other flag)
             while True:
                 try:
                     data = sock.recv(size, socket.MSG_WAITALL)
@@ -222,7 +229,8 @@ def sendData(sock, data):
 _GLOBAL_DEFAULT_TIMEOUT = object()
 
 
-def createSocket(bind=None, connect=None, reuseaddr=False, keepalive=True, timeout=_GLOBAL_DEFAULT_TIMEOUT, noinherit=False, ipv6=False, nodelay=True):
+def createSocket(bind=None, connect=None, reuseaddr=False, keepalive=True,
+                 timeout=_GLOBAL_DEFAULT_TIMEOUT, noinherit=False, ipv6=False, nodelay=True, sslContext=None):
     """
     Create a socket. Default socket options are keepalive and IPv4 family, and nodelay (nagle disabled).
     If 'bind' or 'connect' is a string, it is assumed a Unix domain socket is requested.
@@ -268,6 +276,13 @@ def createSocket(bind=None, connect=None, reuseaddr=False, keepalive=True, timeo
     else:
         raise ValueError("unknown bind or connect format.")
     sock = socket.socket(family, socket.SOCK_STREAM)
+    if sslContext:
+        if bind:
+            sock = sslContext.wrap_socket(sock, server_side=True)
+        elif connect:
+            sock = sslContext.wrap_socket(sock, server_side=False, server_hostname=connect[0])
+        else:
+            sock = sslContext.wrap_socket(sock, server_side=False)
     if nodelay:
         setNoDelay(sock)
     if reuseaddr:
@@ -285,7 +300,7 @@ def createSocket(bind=None, connect=None, reuseaddr=False, keepalive=True, timeo
             sock.bind(bind)
         try:
             sock.listen(100)
-        except Exception:
+        except (OSError, IOError):
             pass
     if connect:
         try:
@@ -425,12 +440,12 @@ except ImportError:
 
 class SocketConnection(object):
     """A wrapper class for plain sockets, containing various methods such as :meth:`send` and :meth:`recv`"""
-    __slots__ = ["sock", "objectId", "pyroInstances"]
-
-    def __init__(self, sock, objectId=None):
+    def __init__(self, sock, objectId=None, keep_open=False):
         self.sock = sock
         self.objectId = objectId
         self.pyroInstances = {}    # pyro objects for instance_mode=session
+        self.tracked_resources = weakref.WeakSet()      # weakrefs to resources for this connection
+        self.keep_open = keep_open
 
     def __del__(self):
         self.close()
@@ -448,6 +463,8 @@ class SocketConnection(object):
         return receiveData(self.sock, size)
 
     def close(self):
+        if self.keep_open:
+            return
         try:
             self.sock.shutdown(socket.SHUT_RDWR)
         except:
@@ -456,7 +473,13 @@ class SocketConnection(object):
             self.sock.close()
         except:
             pass
-        self.pyroInstances = {}   # force releasing the session instances
+        self.pyroInstances.clear()   # release the session instances
+        for rsc in self.tracked_resources:
+            try:
+                rsc.close()     # it is assumed a 'resource' has a close method.
+            except Exception:
+                pass
+        self.tracked_resources.clear()
 
     def fileno(self):
         return self.sock.fileno()
@@ -470,6 +493,12 @@ class SocketConnection(object):
     def getTimeout(self):
         return self.sock.gettimeout()
 
+    def getpeercert(self):
+        try:
+            return self.sock.getpeercert()
+        except AttributeError:
+            return None
+
     timeout = property(getTimeout, setTimeout)
 
 
@@ -479,7 +508,7 @@ def family_str(sock):
         return "IPv4"
     if f == socket.AF_INET6:
         return "IPv6"
-    if f == socket.AF_UNIX:
+    if hasattr(socket, "AF_UNIX") and f == socket.AF_UNIX:
         return "Unix"
     return "???"
 
@@ -536,3 +565,59 @@ def interruptSocket(address):
         sock.close()
     except socket.error:
         pass
+
+
+__ssl_server_context = None
+__ssl_client_context = None
+
+
+def getSSLcontext(servercert="", serverkey="", clientcert="", clientkey="", cacerts="", keypassword=""):
+    """creates an SSL context and caches it, so you have to set the parameters correctly before doing anything"""
+    global __ssl_client_context, __ssl_server_context
+    if not ssl:
+        raise ValueError("SSL requested but ssl module is not available")
+    else:
+        # Theoretically, the SSL support works on python versions older than the ones checked below.
+        # however, a few important security changes were included in these versions
+        # (disabling vulnerable cyphers and protocols by default). So change this at your own peril.
+        if sys.version_info < (2, 7, 11):
+            raise RuntimeError("need Python 2.7.11 or newer to properly use SSL")
+        if (3, 0) < sys.version_info < (3, 4, 4):
+            raise RuntimeError("need Python 3.4.4 or newer to properly use SSL")
+    if servercert:
+        if clientcert:
+            raise ValueError("can't have both server cert and client cert")
+        # server context
+        if __ssl_server_context:
+            return __ssl_server_context
+        if not os.path.isfile(servercert):
+            raise IOError("server cert file not found")
+        if serverkey and not os.path.isfile(serverkey):
+            raise IOError("server key file not found")
+        __ssl_server_context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
+        __ssl_server_context.load_cert_chain(servercert, serverkey or None, keypassword or None)
+        if cacerts:
+            if os.path.isdir(cacerts):
+                __ssl_server_context.load_verify_locations(capath=cacerts)
+            else:
+                __ssl_server_context.load_verify_locations(cafile=cacerts)
+        if config.SSL_REQUIRECLIENTCERT:
+            __ssl_server_context.verify_mode = ssl.CERT_REQUIRED   # 2-way ssl, server+client certs
+        else:
+            __ssl_server_context.verify_mode = ssl.CERT_NONE   # 1-way ssl, server cert only
+        return __ssl_server_context
+    else:
+        # client context
+        if __ssl_client_context:
+            return __ssl_client_context
+        __ssl_client_context = ssl.create_default_context(ssl.Purpose.SERVER_AUTH)
+        if clientcert:
+            if not os.path.isfile(clientcert):
+                raise IOError("client cert file not found")
+            __ssl_client_context.load_cert_chain(clientcert, clientkey or None, keypassword or None)
+        if cacerts:
+            if os.path.isdir(cacerts):
+                __ssl_client_context.load_verify_locations(capath=cacerts)
+            else:
+                __ssl_client_context.load_verify_locations(cafile=cacerts)
+        return __ssl_client_context
